@@ -1,4 +1,4 @@
-# Copyright 2004-2018 Tom Rothamel <pytom@bishoujo.us>
+# Copyright 2004-2021 Tom Rothamel <pytom@bishoujo.us>
 #
 # Permission is hereby granted, free of charge, to any person
 # obtaining a copy of this software and associated documentation files
@@ -26,7 +26,8 @@
 # When updating this file, consider if lint.py or warp.py also need
 # updating.
 
-from __future__ import print_function
+from __future__ import division, absolute_import, with_statement, print_function, unicode_literals
+from renpy.compat import *
 
 import renpy.display
 import renpy.test
@@ -89,7 +90,7 @@ class ParameterInfo(object):
             best job it can.
         """
 
-        values = { }
+        values = renpy.python.RevertableDict()
         rv = { }
 
         if args is None:
@@ -107,7 +108,7 @@ class ParameterInfo(object):
 
         extrapos = tuple(args[len(self.positional):])
 
-        for name, value in kwargs.iteritems():
+        for name, value in kwargs.items():
             if name in values:
                 if not ignore_errors:
                     raise Exception("Parameter %s has two values." % name)
@@ -137,8 +138,12 @@ class ParameterInfo(object):
 
         if self.extrakw:
             rv[self.extrakw] = values
-        elif values and not ignore_errors:
-            raise Exception("Unknown keyword arguments: %s" % ( ", ".join(values.keys())))
+
+        elif values.get("_ignore_extra_kwargs", False):
+            pass
+
+        elif values and (not ignore_errors):
+            raise Exception("Unknown keyword arguments: %s" % (", ".join(list(values.keys()))))
 
         return rv
 
@@ -220,7 +225,7 @@ def __newobj__(cls, *args):
 pyexpr_list = [ ]
 
 
-class PyExpr(unicode):
+class PyExpr(str):
     """
     Represents a string containing python code.
     """
@@ -231,7 +236,7 @@ class PyExpr(unicode):
         ]
 
     def __new__(cls, s, filename, linenumber):
-        self = unicode.__new__(cls, s)
+        self = str.__new__(cls, s)
         self.filename = filename
         self.linenumber = linenumber
 
@@ -242,7 +247,17 @@ class PyExpr(unicode):
         return self
 
     def __getnewargs__(self):
-        return (unicode(self), self.filename, self.linenumber)  # E1101
+        return (str(self), self.filename, self.linenumber) # E1101
+
+
+def probably_side_effect_free(expr):
+    """
+    Returns true if an expr probably does not have side effects, and should
+    be predicted. Basically, this just whitelists a set of characters that
+    doesn't allow for a function call.
+    """
+
+    return not ("(" in expr)
 
 
 class PyCode(object):
@@ -275,7 +290,7 @@ class PyCode(object):
 
         # The time is necessary so we can disambiguate between Python
         # blocks on the same line in different script versions.
-        self.location = loc + ( int(time.time()), )
+        self.location = loc + (int(time.time()),)
         self.mode = mode
 
         # This will be initialized later on, after we are serialized.
@@ -294,14 +309,14 @@ class PyCode(object):
             pass
 
         code = self.source
-        if isinstance(code, renpy.python.ast.AST):  # @UndefinedVariable
-            code = renpy.python.ast.dump(code)  # @UndefinedVariable
+        if isinstance(code, renpy.python.ast.AST): # @UndefinedVariable
+            code = renpy.python.ast.dump(code) # @UndefinedVariable
 
-        self.hash = chr(renpy.bytecode_version) + hashlib.md5(repr(self.location) + code.encode("utf-8")).digest()
+        self.hash = bchr(renpy.bytecode_version) + hashlib.md5((repr(self.location) + code).encode("utf-8")).digest()
         return self.hash
 
 
-def chain_block(block, next):  # @ReservedAssignment
+def chain_block(block, next): # @ReservedAssignment
     """
     This is called to chain together all of the nodes in a block. Node
     n is chained with node n+1, while the last node is chained with
@@ -327,7 +342,7 @@ class Scry(object):
     def __getattr__(self, name):
         return None
 
-    def next(self):  # @ReservedAssignment
+    def next(self): # @ReservedAssignment
         if self._next is None:
             return None
         else:
@@ -342,9 +357,10 @@ class Node(object):
     A node in the abstract syntax tree of the program.
 
     @ivar name: The name of this node.
-
     @ivar filename: The filename where this node comes from.
     @ivar linenumber: The line number of the line on which this node is defined.
+    @ivar next: The statement that will execute after this one.
+    @ivar statement_start: If present, the first node that makes up the statement that includes this node.
     """
 
     __slots__ = [
@@ -352,6 +368,7 @@ class Node(object):
         'filename',
         'linenumber',
         'next',
+        'statement_start',
         ]
 
     # True if this node is translatable, false otherwise. (This can be set on
@@ -388,7 +405,7 @@ class Node(object):
         tuple returned must be hashable.
         """
 
-        return ( id(self), )
+        return (id(self),)
 
     def get_children(self, f):
         """
@@ -411,7 +428,7 @@ class Node(object):
     # get_init is only present on statements that define it.
     get_init = None
 
-    def chain(self, next):  # @ReservedAssignment
+    def chain(self, next): # @ReservedAssignment
         """
         This is called with the Node node that should be followed after
         executing this node, and all nodes that this node
@@ -450,8 +467,8 @@ class Node(object):
 
     def predict(self):
         """
-        This is called to predictively load images from this node.  It
-        should cause renpy.display.predict.image and
+        This is called to predictively load images from this node. It
+        should cause renpy.display.predict.displayable and
         renpy.display.predict.screen to be called as necessary.
         """
 
@@ -467,7 +484,7 @@ class Node(object):
         """
 
         rv = Scry()
-        rv._next = self.next  # W0201
+        rv._next = self.next # W0201
         return rv
 
     def restructure(self, callback):
@@ -519,7 +536,7 @@ def say_menu_with(expression, callback):
 
     if expression is not None:
         what = renpy.python.py_eval(expression)
-    elif renpy.store.default_transition and renpy.game.preferences.transitions == 2:
+    elif renpy.store.default_transition and renpy.game.preferences.transitions == 2: # @UndefinedVariable
         what = renpy.store.default_transition
     else:
         return
@@ -527,14 +544,11 @@ def say_menu_with(expression, callback):
     if not what:
         return
 
-    if renpy.game.preferences.transitions:
+    if renpy.game.preferences.transitions: # @UndefinedVariable
         # renpy.game.interface.set_transition(what)
         callback(what)
 
 
-fast_who_pattern = re.compile(r'[a-zA-Z_][a-zA-Z_0-9]*$')
-
-
 def eval_who(who, fast=None):
     """
     Evaluates the `who` parameter to a say statement.
@@ -543,25 +557,21 @@ def eval_who(who, fast=None):
     if who is None:
         return None
 
-    if fast is None:
-        fast = bool(fast_who_pattern.match(who))
+    if 'store.character' in renpy.python.store_dicts:
+        rv = renpy.python.store_dicts['store.character'].get(who, None)
+    else:
+        rv = None
 
-    if fast:
-
-        if 'store.character' in renpy.python.store_dicts:
-            rv = renpy.python.store_dicts['store.character'].get(who, None)
-        else:
-            rv = None
-
-        if rv is None:
-            rv = renpy.python.store_dicts['store'].get(who, None)
-
-        if rv is None:
-            raise Exception("Sayer '%s' is not defined." % who.encode("utf-8"))
+    if rv is None:
+        rv = renpy.python.store_dicts['store'].get(who, None)
 
+    if rv is not None:
         return rv
 
-    return renpy.python.py_eval(who)
+    try:
+        return renpy.python.py_eval(who)
+    except:
+        raise Exception("Sayer '%s' is not defined." % who)
 
 
 class Say(Node):
@@ -574,6 +584,9 @@ class Say(Node):
         'interact',
         'attributes',
         'arguments',
+        'temporary_attributes',
+        'rollback',
+        'identifier',
         ]
 
     def diff_info(self):
@@ -584,9 +597,11 @@ class Say(Node):
         self.attributes = None
         self.interact = True
         self.arguments = None
+        self.temporary_attributes = None
+        self.rollback = "normal"
         return self
 
-    def __init__(self, loc, who, what, with_, interact=True, attributes=None, arguments=None):
+    def __init__(self, loc, who, what, with_, interact=True, attributes=None, arguments=None, temporary_attributes=None, identifier=None):
 
         super(Say, self).__init__(loc)
 
@@ -611,6 +626,13 @@ class Say(Node):
         # speaking, or None to disable this behavior.
         self.attributes = attributes
 
+        # Ditto for temporary attributes.
+        self.temporary_attributes = temporary_attributes
+
+        # If given, write in the identifier.
+        if identifier is not None:
+            self.identifier = identifier
+
     def get_code(self, dialogue_filter=None):
         rv = [ ]
 
@@ -620,15 +642,23 @@ class Say(Node):
         if self.attributes is not None:
             rv.extend(self.attributes)
 
+        if self.temporary_attributes:
+            rv.append("@")
+            rv.extend(self.temporary_attributes)
+
         what = self.what
         if dialogue_filter is not None:
             what = dialogue_filter(what)
 
-        rv.append(renpy.translation.encode_say_string(what))
+        rv.append(renpy.translation.encode_say_string(what)) # @UndefinedVariable
 
         if not self.interact:
             rv.append("nointeract")
 
+        if getattr(self, "identifier", None):
+            rv.append("id")
+            rv.append(getattr(self, "identifier", None))
+
         if self.with_:
             rv.append("with")
             rv.append(self.with_)
@@ -641,24 +671,29 @@ class Say(Node):
     def execute(self):
 
         next_node(self.next)
-        statement_name("say")
 
         try:
 
             renpy.game.context().say_attributes = self.attributes
+            renpy.game.context().temporary_attributes = self.temporary_attributes
 
             who = eval_who(self.who, self.who_fast)
 
+            if who is not None:
+                statement_name(getattr(who, "statement_name", "say"))
+            else:
+                statement_name("say")
+
             if not (
                     (who is None) or
                     callable(who) or
-                    isinstance(who, basestring) ):
+                    isinstance(who, basestring)):
 
                 raise Exception("Sayer %s is not a function or string." % self.who.encode("utf-8"))
 
             what = self.what
             if renpy.config.say_menu_text_filter:
-                what = renpy.config.say_menu_text_filter(what)  # E1102
+                what = renpy.config.say_menu_text_filter(what) # E1102
 
             renpy.store._last_raw_what = what
 
@@ -668,6 +703,8 @@ class Say(Node):
                 args = tuple()
                 kwargs = dict()
 
+            kwargs.setdefault("interact", self.interact)
+
             if getattr(who, "record_say", True):
                 renpy.store._last_say_who = self.who
                 renpy.store._last_say_what = what
@@ -675,18 +712,21 @@ class Say(Node):
                 renpy.store._last_say_kwargs = kwargs
 
             say_menu_with(self.with_, renpy.game.interface.set_transition)
-            renpy.exports.say(who, what, interact=self.interact, *args, **kwargs)
+            renpy.exports.say(who, what, *args, **kwargs)
 
         finally:
             renpy.game.context().say_attributes = None
+            renpy.game.context().temporary_attributes = None
 
     def predict(self):
 
         old_attributes = renpy.game.context().say_attributes
+        old_temporary_attributes = renpy.game.context().temporary_attributes
 
         try:
 
             renpy.game.context().say_attributes = self.attributes
+            renpy.game.context().temporary_attributes = self.temporary_attributes
 
             who = eval_who(self.who, self.who_fast)
 
@@ -703,6 +743,7 @@ class Say(Node):
 
         finally:
             renpy.game.context().say_attributes = old_attributes
+            renpy.game.context().temporary_attributes = old_temporary_attributes
 
         return [ self.next ]
 
@@ -720,7 +761,7 @@ class Say(Node):
 
 
 # Copy the descriptor.
-setattr(Say, "with", Say.with_)  # E1101
+setattr(Say, "with", Say.with_) # E1101
 
 
 class Init(Node):
@@ -748,7 +789,7 @@ class Init(Node):
     # We handle chaining specially. We want to chain together the nodes in
     # the block, but we want that chain to end in None, and we also want
     # this node to just continue on to the next node in normal execution.
-    def chain(self, next):  # @ReservedAssignment
+    def chain(self, next): # @ReservedAssignment
         self.next = next
 
         chain_block(self.block, None)
@@ -767,7 +808,6 @@ class Label(Node):
     rollback = "force"
 
     translation_relevant = True
-
     __slots__ = [
         'name',
         'parameters',
@@ -806,7 +846,7 @@ class Label(Node):
         for i in self.block:
             i.get_children(f)
 
-    def chain(self, next):  # @ReservedAssignment
+    def chain(self, next): # @ReservedAssignment
 
         if self.block:
             self.next = self.block[0]
@@ -822,7 +862,7 @@ class Label(Node):
 
         values = apply_arguments(self.parameters, renpy.store._args, renpy.store._kwargs)
 
-        for k, v in values.iteritems():
+        for k, v in values.items():
             renpy.exports.dynamic(k)
             setattr(renpy.store, k, v)
 
@@ -1294,7 +1334,7 @@ class Hide(Node):
 
         layer = renpy.exports.default_layer(layer, tag)
 
-        renpy.game.context().images.predict_hide(tag, layer)
+        renpy.game.context().images.predict_hide(layer, (tag,))
 
         return [ self.next ]
 
@@ -1409,13 +1449,26 @@ class Call(Node):
             args, kwargs = self.arguments.evaluate()
             renpy.store._args = args
             renpy.store._kwargs = kwargs
+        else:
+            renpy.store._args = None
+            renpy.store._kwargs = None
 
     def predict(self):
 
         label = self.label
 
         if self.expression:
-            label = renpy.python.py_eval(label)
+
+            if not probably_side_effect_free(label):
+                return [ ]
+
+            try:
+                label = renpy.python.py_eval(label)
+            except:
+                return [ ]
+
+            if not renpy.game.script.has_label(label):
+                return [ ]
 
         return [ renpy.game.context().predict_call(label, self.next.name) ]
 
@@ -1439,10 +1492,10 @@ class Return(Node):
         self.expression = expression
 
     def diff_info(self):
-        return (Return, )
+        return (Return,)
 
     # We don't care what the next node is.
-    def chain(self, next):  # @ReservedAssignment
+    def chain(self, next): # @ReservedAssignment
         self.next = None
         return
 
@@ -1486,14 +1539,29 @@ class Menu(Node):
         'items',
         'set',
         'with_',
+        'has_caption',
+        'arguments',
+        'item_arguments',
+        'rollback',
         ]
 
-    def __init__(self, loc, items, set, with_):  # @ReservedAssignment
+    def __new__(cls, *args, **kwargs):
+        self = Node.__new__(cls)
+        self.has_caption = False
+        self.arguments = None
+        self.item_arguments = None
+        self.rollback = "force"
+        return self
+
+    def __init__(self, loc, items, set, with_, has_caption, arguments, item_arguments): # @ReservedAssignment
         super(Menu, self).__init__(loc)
 
         self.items = items
         self.set = set
         self.with_ = with_
+        self.has_caption = has_caption
+        self.arguments = arguments
+        self.item_arguments = item_arguments
 
     def diff_info(self):
         return (Menu,)
@@ -1507,7 +1575,7 @@ class Menu(Node):
                     i.get_children(f)
 
     # Blocks of statements in a choice continue after the menu.
-    def chain(self, next):  # @ReservedAssignment
+    def chain(self, next): # @ReservedAssignment
 
         self.next = next
 
@@ -1525,30 +1593,53 @@ class Menu(Node):
     def execute(self):
 
         next_node(self.next)
-        statement_name("menu")
+
+        if self.has_caption:
+            statement_name("menu-with-caption")
+        else:
+            statement_name("menu")
+
+        if self.arguments is not None:
+            args, kwargs = self.arguments.evaluate()
+        else:
+            args = kwargs = None
 
         choices = [ ]
         narration = [ ]
+        item_arguments = [ ]
 
         for i, (label, condition, block) in enumerate(self.items):
 
             if renpy.config.say_menu_text_filter:
                 label = renpy.config.say_menu_text_filter(label)
 
+            has_item = False
+
             if block is None:
                 if renpy.config.narrator_menu and label:
                     narration.append(label)
                 else:
                     choices.append((label, condition, None))
+                    has_item = True
+
             else:
                 choices.append((label, condition, i))
+                has_item = True
+
                 next_node(block[0])
 
+            if has_item:
+                if self.item_arguments and (self.item_arguments[i] is not None):
+                    item_arguments.append(self.item_arguments[i].evaluate())
+                else:
+                    item_arguments.append((tuple(), dict()))
+
         if narration:
             renpy.exports.say(None, "\n".join(narration), interact=False)
 
         say_menu_with(self.with_, renpy.game.interface.set_transition)
-        choice = renpy.exports.menu(choices, self.set)
+
+        choice = renpy.exports.menu(choices, self.set, args, kwargs, item_arguments)
 
         if choice is not None:
             next_node(self.items[choice][2][0])
@@ -1583,7 +1674,7 @@ class Menu(Node):
                 callback(block)
 
 
-setattr(Menu, "with", Menu.with_)  # E1101
+setattr(Menu, "with", Menu.with_) # E1101
 
 
 # Goto is considered harmful. So we decided to name it "jump"
@@ -1605,7 +1696,7 @@ class Jump(Node):
         return (Jump, self.target, self.expression)
 
     # We don't care what our next node is.
-    def chain(self, next):  # @ReservedAssignment
+    def chain(self, next): # @ReservedAssignment
         self.next = None
         return
 
@@ -1624,10 +1715,22 @@ class Jump(Node):
 
     def predict(self):
 
+        label = self.target
+
         if self.expression:
-            return [ ]
-        else:
-            return [ renpy.game.script.lookup(self.target) ]
+
+            if not probably_side_effect_free(label):
+                return [ ]
+
+            try:
+                label = renpy.python.py_eval(label)
+            except:
+                return [ ]
+
+            if not renpy.game.script.has_label(label):
+                return [ ]
+
+        return [ renpy.game.script.lookup(label) ]
 
     def scry(self):
         rv = Node.scry(self)
@@ -1674,7 +1777,7 @@ class While(Node):
         for i in self.block:
             i.get_children(f)
 
-    def chain(self, next):  # @ReservedAssignment
+    def chain(self, next): # @ReservedAssignment
         self.next = next
         chain_block(self.block, self)
 
@@ -1727,7 +1830,7 @@ class If(Node):
             for i in block:
                 i.get_children(f)
 
-    def chain(self, next):  # @ReservedAssignment
+    def chain(self, next): # @ReservedAssignment
         self.next = next
 
         for _condition, block in self.entries:
@@ -1773,6 +1876,9 @@ class UserStatement(Node):
         'block',
         'translatable',
         'code_block',
+        'translation_relevant',
+        'rollback',
+        'subparses',
         ]
 
     def __new__(cls, *args, **kwargs):
@@ -1780,20 +1886,22 @@ class UserStatement(Node):
         self.block = [ ]
         self.code_block = None
         self.translatable = False
+        self.translation_relevant = False
+        self.rollback = "normal"
+        self.subparses = [ ]
         return self
 
-    def __init__(self, loc, line, block):
+    def __init__(self, loc, line, block, parsed):
 
         super(UserStatement, self).__init__(loc)
+        self.code_block = None
+        self.parsed = parsed
         self.line = line
         self.block = block
-        self.code_block = None
-        self.parsed = None
+        self.subparses = [ ]
 
         self.name = self.call("label")
-
-        # Do not store the parse.
-        self.parsed = None
+        self.rollback = renpy.statements.get("rollback", self.parsed) or "normal"
 
     def __repr__(self):
         return "<UserStatement {!r}>".format(self.line)
@@ -1801,34 +1909,47 @@ class UserStatement(Node):
     def get_children(self, f):
         f(self)
 
-        if not self.code_block:
-            return
+        if self.code_block is not None:
+            for i in self.code_block:
+                i.get_children(f)
 
-        for i in self.code_block:
-            i.get_children(f)
+        for i in self.subparses:
+            for j in i.block:
+                j.get_children(f)
 
-    def chain(self, next):  # @ReservedAssignment
+    def chain(self, next): # @ReservedAssignment
         self.next = next
 
         if self.code_block is not None:
             chain_block(self.code_block, next)
 
+        for i in self.subparses:
+            chain_block(i.block, next)
+
     def replace_next(self, old, new):
         Node.replace_next(self, old, new)
 
         if (self.code_block) and (self.code_block[0] is old):
             self.code_block.insert(0, new)
 
+        for i in self.subparses:
+            if i.block[0] is old:
+                self.code_block.insert(0, new)
+
     def restructure(self, callback):
         if self.code_block:
             callback(self.code_block)
 
+        for i in self.subparses:
+            callback(i.block)
+
     def diff_info(self):
         return (UserStatement, self.line)
 
     def call(self, method, *args, **kwargs):
 
         parsed = self.parsed
+
         if parsed is None:
             parsed = renpy.statements.parse(self, self.line, self.block)
             self.parsed = parsed
@@ -1854,10 +1975,25 @@ class UserStatement(Node):
             for i in predictions:
                 renpy.easy.predict(i)
 
-        return [ self.get_next() ]
+        if self.parsed and renpy.statements.get("predict_all", self.parsed):
+            return [ i.block[0] for i in self.subparses ] + [ self.next ]
+
+        if self.next:
+            next_label = self.next.name
+        else:
+            next_label = None
+
+        next_list = self.call("predict_next", next_label)
+
+        if next_list is not None:
+            nexts = [ renpy.game.script.lookup_or_none(i) for i in next_list if i is not None ]
+            return [ i for i in nexts if i is not None ]
+
+        return [ self.next ]
 
     def get_name(self):
         parsed = self.parsed
+
         if parsed is None:
             parsed = renpy.statements.parse(self, self.line, self.block)
             self.parsed = parsed
@@ -1893,6 +2029,32 @@ class UserStatement(Node):
         return False
 
 
+class PostUserStatement(Node):
+
+    __slots__ = [
+        'parent',
+        ]
+
+    def __init__(self, loc, parent):
+
+        super(PostUserStatement, self).__init__(loc)
+        self.parent = parent
+
+        self.name = self.parent.call("post_label")
+
+    def __repr__(self):
+        return "<PostUserStatement {!r}>".format(self.parent.line)
+
+    def diff_info(self):
+        return (PostUserStatement, self.parent.line)
+
+    def execute(self):
+        next_node(self.next)
+        statement_name("post " + self.parent.get_name())
+
+        self.parent.call("post_execute")
+
+
 def create_store(name):
     if name not in renpy.config.special_namespaces:
         renpy.python.create_store(name)
@@ -1906,6 +2068,9 @@ class StoreNamespace(object):
     def set(self, name, value):
         renpy.python.store_dicts[self.store][name] = value
 
+    def get(self, name):
+        return renpy.python.store_dicts[self.store][name]
+
 
 def get_namespace(store):
     """
@@ -1921,7 +2086,14 @@ def get_namespace(store):
 
 # Config variables that are set twice - once when the rpy is first loaded,
 # and then again at init time.
-EARLY_CONFIG = { "save_directory" }
+EARLY_CONFIG = {
+    "save_directory",
+    "allow_duplicate_labels",
+    "keyword_after_python",
+    "steam_appid",
+    "name",
+    "version",
+}
 
 define_statements = [ ]
 
@@ -1932,18 +2104,29 @@ class Define(Node):
         'varname',
         'code',
         'store',
+        'operator',
+        'index',
         ]
 
     def __new__(cls, *args, **kwargs):
         self = Node.__new__(cls)
         self.store = 'store'
+        self.operator = '='
+        self.index = None
         return self
 
-    def __init__(self, loc, store, name, expr):
+    def __init__(self, loc, store, name, index, operator, expr):
         super(Define, self).__init__(loc)
 
         self.store = store
         self.varname = name
+
+        if index is not None:
+            self.index = PyCode(index, loc=loc, mode='eval')
+        else:
+            self.index = None
+
+        self.operator = operator
         self.code = PyCode(expr, loc=loc, mode='eval')
 
     def diff_info(self):
@@ -1952,7 +2135,14 @@ class Define(Node):
     def early_execute(self):
         create_store(self.store)
 
+        if self.operator != "=":
+            return
+
+        if self.index is not None:
+            return
+
         if self.store == "store.config" and self.varname in EARLY_CONFIG:
+
             value = renpy.python.py_eval_bytecode(self.code.bytecode)
             setattr(renpy.config, self.varname, value)
 
@@ -1963,25 +2153,52 @@ class Define(Node):
 
         define_statements.append(self)
 
-        value = renpy.python.py_eval_bytecode(self.code.bytecode)
-
         if self.store == 'store':
-            renpy.exports.pure(self.varname)
             renpy.dump.definitions.append((self.varname, self.filename, self.linenumber))
         else:
             renpy.dump.definitions.append((self.store[6:] + "." + self.varname, self.filename, self.linenumber))
 
-        ns, _special = get_namespace(self.store)
-        ns.set(self.varname, value)
+        if self.operator == "=" and self.index is None:
+            renpy.exports.pure(self.store + "." + self.varname)
+
+        self.set()
 
     def redefine(self, stores):
 
         if self.store not in stores:
             return
 
+        self.set()
+
+    def set(self):
+
         value = renpy.python.py_eval_bytecode(self.code.bytecode)
         ns, _special = get_namespace(self.store)
-        ns.set(self.varname, value)
+
+        if (self.index is None) and (self.operator == "="):
+            ns.set(self.varname, value)
+            return
+
+        base = ns.get(self.varname)
+        old = base
+
+        if self.index:
+            key = renpy.python.py_eval_bytecode(self.index.bytecode)
+
+            if self.operator != "=":
+                old = base[key]
+
+        if self.operator == "=":
+            new = value
+        elif self.operator == "+=":
+            new = old + value
+        elif self.operator == "|=":
+            new = old | value
+
+        if self.index:
+            base[key] = new
+        else:
+            ns.set(self.varname, new)
 
 
 def redefine(stores):
@@ -1993,7 +2210,6 @@ def redefine(stores):
         i.redefine(stores)
 
 
-
 # All the default statements, in the order they were registered.
 default_statements = [ ]
 
@@ -2080,10 +2296,9 @@ class Screen(Node):
 
     def __init__(self, loc, screen):
         """
-        @param name: The name of the image being defined.
-
-        @param expr: An expression yielding a Displayable that is
-        assigned to the image.
+        @param screen: The screen object being defined.
+        In SL1, an instance of screenlang.ScreenLangScreen.
+        In SL2, an instance of sl2.slast.SLScreen.
         """
 
         super(Screen, self).__init__(loc)
@@ -2100,11 +2315,11 @@ class Screen(Node):
         self.screen.define((self.filename, self.linenumber))
         renpy.dump.screens.append((self.screen.name, self.filename, self.linenumber))
 
-
 ################################################################################
 # Translations
 ################################################################################
 
+
 class Translate(Node):
     """
     A translation block, produced either by explicit translation statements
@@ -2142,7 +2357,7 @@ class Translate(Node):
     def diff_info(self):
         return (Translate, self.identifier, self.language)
 
-    def chain(self, next):  # @ReservedAssignment
+    def chain(self, next): # @ReservedAssignment
         if self.block:
             self.next = self.block[0]
             chain_block(self.block, next)
@@ -2171,8 +2386,8 @@ class Translate(Node):
             next_node(self.next)
             raise Exception("Translation nodes cannot be run directly.")
 
-        if self.identifier not in renpy.game.persistent._seen_translates:  # @UndefinedVariable
-            renpy.game.persistent._seen_translates.add(self.identifier)  # @UndefinedVariable
+        if self.identifier not in renpy.game.persistent._seen_translates: # @UndefinedVariable
+            renpy.game.persistent._seen_translates.add(self.identifier) # @UndefinedVariable
             renpy.game.seen_translates_count += 1
             renpy.game.new_translates_count += 1
 
@@ -2321,7 +2536,7 @@ class TranslateBlock(Node):
     # We handle chaining specially. We want to chain together the nodes in
     # the block, but we want that chain to end in None, and we also want
     # this node to just continue on to the next node in normal execution.
-    def chain(self, next):  # @ReservedAssignment
+    def chain(self, next): # @ReservedAssignment
         self.next = next
         chain_block(self.block, None)
 
@@ -2389,7 +2604,7 @@ class Style(Node):
             if not renpy.exports.variant(variant):
                 return
 
-        s = renpy.style.get_or_create_style(self.style_name)  # @UndefinedVariable
+        s = renpy.style.get_or_create_style(self.style_name) # @UndefinedVariable
 
         if self.clear:
             s.clear()
@@ -2449,3 +2664,26 @@ class Testcase(Node):
         statement_name("testcase")
 
         renpy.test.testexecution.testcases[self.label] = self.test
+
+
+class RPY(Node):
+    __slots__ = [
+        "rest"
+        ]
+
+    def __init__(self, loc, rest):
+        super(RPY, self).__init__(loc)
+
+        self.rest = rest
+
+    def diff_info(self):
+        return (RPY, self.rest)
+
+    def execute(self):
+        next_node(self.next)
+        statement_name("rpy")
+
+        # rpy python 3 is run in Script.finish_load.
+
+    def get_code(self):
+        return "rpy " + " ".join(self.rest)
