@@ -1,4 +1,4 @@
-# Copyright 2004-2019 Tom Rothamel <pytom@bishoujo.us>
+# Copyright 2004-2021 Tom Rothamel <pytom@bishoujo.us>
 #
 # Permission is hereby granted, free of charge, to any person
 # obtaining a copy of this software and associated documentation files
@@ -21,16 +21,18 @@
 
 # The Character object (and friends).
 
-from __future__ import print_function
+from __future__ import division, absolute_import, with_statement, print_function, unicode_literals
+from renpy.compat import *
 
 import renpy.display
 
 import re
 import os
 import collections
+import renpy.six as six
 
 # This matches the dialogue-relevant text tags.
-TAG_RE = re.compile(r'(\{\{)|(\{(p|w|nw|fast)(?:\=([^}]*))?\})', re.S)
+TAG_RE = re.compile(r'(\{\{)|(\{(p|w|nw|fast|done)(?:\=([^}]*))?\})', re.S)
 
 less_pauses = ("RENPY_LESS_PAUSES" in os.environ)
 
@@ -58,17 +60,20 @@ class DialogueTextTags(object):
         # True if we've encountered the no-wait tag.
         self.no_wait = False
 
+        # Does this statement have a done tag?
+        self.has_done = False
+
         i = iter(TAG_RE.split(s))
 
         while True:
 
             try:
-                self.text += i.next()
+                self.text += next(i)
 
-                quoted = i.next()
-                full_tag = i.next()
-                tag = i.next()
-                value = i.next()
+                quoted = next(i)
+                full_tag = next(i)
+                tag = next(i)
+                value = next(i)
 
                 if value is not None:
                     value = float(value)
@@ -92,6 +97,11 @@ class DialogueTextTags(object):
                     self.pause_delay = [ ]
                     self.no_wait = False
 
+                elif tag == "done":
+                    self.has_done = True
+                    self.text += full_tag
+                    break
+
                 self.text += full_tag
 
             except StopIteration:
@@ -99,6 +109,25 @@ class DialogueTextTags(object):
 
         self.pause_end.append(len(self.text))
 
+        while True:
+
+            try:
+                self.text += next(i)
+
+                quoted = next(i)
+                full_tag = next(i)
+                tag = next(i)
+                value = next(i)
+
+                if quoted is not None:
+                    self.text += quoted
+                    continue
+
+                self.text += full_tag
+
+            except StopIteration:
+                break
+
         if self.no_wait:
             self.pause_delay.append(0)
         else:
@@ -343,7 +372,7 @@ class SlowDone(object):
     delay = None
     ctc_kwargs = { }
 
-    def __init__(self, ctc, ctc_position, callback, interact, type, cb_args, delay, ctc_kwargs):  # @ReservedAssignment
+    def __init__(self, ctc, ctc_position, callback, interact, type, cb_args, delay, ctc_kwargs): # @ReservedAssignment
         self.ctc = ctc
         self.ctc_position = ctc_position
         self.callback = callback
@@ -378,7 +407,6 @@ class SlowDone(object):
         for c in self.callback:
             c("slow_done", interact=self.interact, type=self.type, **self.cb_args)
 
-
 # This function takes care of repeatably showing the screen as part of
 # an interaction.
 
@@ -397,12 +425,13 @@ def display_say(
         cb_args,
         with_none,
         callback,
-        type,  # @ReservedAssignment
+        type, # @ReservedAssignment
         checkpoint=True,
         ctc_timedpause=None,
         ctc_force=False,
         advance=True,
-        multiple=None):
+        multiple=None,
+        dtt=None):
 
     # Final is true if this statement should perform an interaction.
 
@@ -473,11 +502,12 @@ def display_say(
     if not interact or renpy.game.preferences.self_voicing:
         all_at_once = True
 
-    dtt = DialogueTextTags(what)
+    if dtt is None:
+        dtt = DialogueTextTags(what)
 
     if all_at_once:
         pause_start = [ dtt.pause_start[0] ]
-        pause_end = [ len(dtt.text) ]
+        pause_end = [ dtt.pause_end[-1] ]
         pause_delay = [ dtt.pause_delay[-1] ]
     else:
         pause_start = dtt.pause_start
@@ -551,7 +581,7 @@ def display_say(
                 if isinstance(what_text, tuple):
                     what_text = renpy.display.screen.get_widget(what_text[0], what_text[1], what_text[2])
 
-                if not isinstance(what_text, renpy.text.text.Text):  # @UndefinedVariable
+                if not isinstance(what_text, renpy.text.text.Text): # @UndefinedVariable
                     raise Exception("The say screen (or show_function) must return a Text object.")
 
                 if what_ctc:
@@ -653,7 +683,6 @@ class HistoryEntry(renpy.object.Object):
 # This is used to flag values that haven't been set by the user.
 NotSet = renpy.object.Sentinel("NotSet")
 
-
 # The number of multiple characters we've seen during the current
 # interaction.
 multiple_count = 0
@@ -854,10 +883,10 @@ class ADVCharacter(object):
             properties=self.properties,
             **self.show_args)
 
-    def resolve_say_attributes(self, predict, attrs, wanted=[], remove=[], skip_trans=False):
+    def resolve_say_attributes(self, predict, attrs, wanted=[], remove=[]):
         """
         Deals with image attributes associated with the current say
-        statement.
+        statement. Returns True if an image is shown, None otherwise.
         """
 
         if not (attrs or wanted or remove):
@@ -874,7 +903,7 @@ class ADVCharacter(object):
         else:
             attrs = tuple(attrs)
 
-        tagged_attrs = ( self.image_tag,) + attrs
+        tagged_attrs = (self.image_tag,) + attrs
         images = renpy.game.context().images
 
         layer = renpy.config.tag_layer.get(self.image_tag, "master")
@@ -890,28 +919,13 @@ class ADVCharacter(object):
             if images.showing(layer, new_image, exact=True):
                 return
 
-            show_image = (self.image_tag,) + attrs + tuple(wanted) + tuple( "-" + i for i in remove)
+            show_image = (self.image_tag,) + attrs + tuple(wanted) + tuple("-" + i for i in remove)
 
             if predict:
-                images.predict_show(new_image)
-
+                renpy.exports.predict_show(new_image)
             else:
-
-                trans = renpy.config.say_attribute_transition
-                layer = renpy.config.say_attribute_transition_layer
-
-                if not skip_trans:
-                    if (trans is not None) and (layer is not None):
-                        renpy.exports.with_statement(None)
-
                 renpy.exports.show(show_image)
-
-                if not skip_trans:
-                    if trans is not None:
-                        if layer is None:
-                            renpy.exports.with_statement(trans)
-                        else:
-                            renpy.exports.transition(trans, layer=layer)
+                return True
 
         else:
 
@@ -950,22 +964,51 @@ class ADVCharacter(object):
             if renpy.config.speaking_attribute is not None:
                 temporary_attrs.insert(0, renpy.config.speaking_attribute)
 
-        self.resolve_say_attributes(predicting, attrs, skip_trans=temporary_attrs)
+        images = renpy.game.context().images
+        before = images.get_attributes(None, self.image_tag)
+        mode = None
+
+        if self.resolve_say_attributes(predicting, attrs):
+            mode = 'permanent'
 
         # This is so late to give resolve_say_attributes time to do some
         # error handling.
         if not self.image_tag:
             return None
 
-        if not temporary_attrs:
-            return None
+        if temporary_attrs:
+            attrs = images.get_attributes(None, self.image_tag)
 
-        images = renpy.game.context().images
-        attrs = images.get_attributes(None, self.image_tag)
+            if self.resolve_say_attributes(predicting, temporary_attrs):
+                mode = 'both' if mode else 'temporary'
 
-        self.resolve_say_attributes(predicting, temporary_attrs)
+        if mode:
+            after = images.get_attributes(None, self.image_tag)
+            self.handle_say_transition(mode, before, after)
 
-        return (attrs, images)
+        if temporary_attrs:
+            return (attrs, images)
+
+    def handle_say_transition(self, mode, before, after):
+
+        before = set(before)
+        after = set(after)
+
+        if before == after:
+            return
+
+        if renpy.config.say_attribute_transition_callback_attrs:
+            delta = (before, after)
+        else:
+            delta = ()
+
+        trans, layer = renpy.config.say_attribute_transition_callback(
+            self.image_tag, mode, *delta)
+        if trans is not None:
+            if layer is None:
+                renpy.exports.with_statement(trans)
+            else:
+                renpy.exports.transition(trans, layer=layer)
 
     def restore_say_attributes(self, predicting, state, interact):
 
@@ -991,25 +1034,10 @@ class ADVCharacter(object):
         if images.showing(None, (self.image_tag,)):
 
             if not predicting:
-
-                trans = renpy.config.say_attribute_transition
-                layer = renpy.config.say_attribute_transition_layer
-
-                if interact:
-                    if (trans is not None) and (layer is not None):
-                        renpy.exports.with_statement(None)
-
                 renpy.exports.show(image_with_attrs)
-
-                if interact:
-                    if trans is not None:
-                        if layer is None:
-                            renpy.exports.with_statement(trans)
-                        else:
-                            renpy.exports.transition(trans, layer=layer)
-
+                return True
             else:
-                images.predict_show(None, image_with_attrs)
+                renpy.exports.predict_show(image_with_attrs)
 
         else:
             images.predict_show(None, image_with_attrs, show=False)
@@ -1019,15 +1047,32 @@ class ADVCharacter(object):
         who = self.name
 
         if self.dynamic:
-            who = renpy.python.py_eval(who)
+            if callable(who):
+                who = who()
+            else:
+                who = renpy.python.py_eval(who)
 
         return renpy.substitutions.substitute(who)[0]
 
     def __str__(self):
-        return unicode(self).encode("utf-8")
+
+        who = self.name
+
+        if self.dynamic:
+            if callable(who):
+                who = who()
+            else:
+                who = renpy.python.py_eval(who)
+
+        rv = renpy.substitutions.substitute(who)[0]
+
+        if PY2:
+            rv = rv.encode("utf-8")
+
+        return rv
 
     def __format__(self, spec):
-        return format(unicode(self), spec)
+        return format(str(self), spec)
 
     def __repr__(self):
         return "<Character: {!r}>".format(self.name)
@@ -1039,10 +1084,28 @@ class ADVCharacter(object):
 
         self("", interact=False, _call_done=False)
 
+    def has_character_arguments(self, **kwargs):
+        """
+        Returns True if `kwargs` contains any keyword arguments that will
+        cause the creation of a new Character object and the proxying of a
+        call to that Character object, and False otherwise.
+        """
+
+        safe_kwargs_keys = { "interact", "_mode", "_call_done", "multiple", "_with_none" }
+
+        for i in kwargs:
+            if i not in safe_kwargs_keys:
+                return False
+
+        return True
+
     def __call__(self, what, interact=True, _call_done=True, multiple=None, **kwargs):
 
+        _mode = kwargs.pop("_mode", None)
+        _with_none = kwargs.pop("_with_none", None)
+
         if kwargs:
-            return Character(kind=self, **kwargs)(what, interact=interact, _call_done=_call_done, multiple=multiple)
+            return Character(kind=self, **kwargs)(what, interact=interact, _call_done=_call_done, multiple=multiple, _mode=_mode, _with_none=_with_none)
 
         # Check self.condition to see if we should show this line at all.
         if not (self.condition is None or renpy.python.py_eval(self.condition)):
@@ -1088,7 +1151,10 @@ class ADVCharacter(object):
         try:
 
             if interact:
-                renpy.exports.mode(self.mode)
+                mode = _mode or self.mode
+                renpy.exports.mode(mode)
+            else:
+                renpy.game.context().deferred_translate_identifier = renpy.game.context().translate_identifier
 
             # Figure out the arguments to display.
             display_args = self.display_args.copy()
@@ -1097,24 +1163,30 @@ class ADVCharacter(object):
             if multiple is not None:
                 display_args["multiple"] = multiple
 
+            if _with_none is not None:
+                display_args["with_none"] = _with_none
+
             who = self.name
 
             # If dynamic is set, evaluate the name expression.
             if self.dynamic:
-                who = renpy.python.py_eval(who)
+                if callable(who):
+                    who = who()
+                else:
+                    who = renpy.python.py_eval(who)
 
             def sub(s, scope=None, force=False, translate=True):
                 return renpy.substitutions.substitute(s, scope=scope, force=force, translate=translate)[0]
 
             if who is not None:
                 if renpy.config.new_substitutions:
-                    who_pattern = sub(self.who_prefix + "[[who]" + self.who_suffix)
+                    who_pattern = sub(sub(self.who_prefix) + "[[who]" + sub(self.who_suffix))
                     who = who_pattern.replace("[who]", sub(who))
                 else:
                     who = self.who_prefix + who + self.who_suffix
 
             if renpy.config.new_substitutions:
-                what_pattern = sub(self.what_prefix + "[[what]" + self.what_suffix)
+                what_pattern = sub(sub(self.what_prefix) + "[[what]" + sub(self.what_suffix))
                 what = what_pattern.replace("[what]", sub(what, translate=True))
             else:
                 what = self.what_prefix + what + self.what_suffix
@@ -1127,11 +1199,13 @@ class ADVCharacter(object):
             else:
                 self.do_add(who, what)
 
+            dtt = DialogueTextTags(what)
+
             # Now, display the damned thing.
-            self.do_display(who, what, cb_args=self.cb_args, **display_args)
+            self.do_display(who, what, cb_args=self.cb_args, dtt=dtt, **display_args)
 
             # Indicate that we're done.
-            if _call_done:
+            if _call_done and not dtt.has_done:
 
                 if multiple is not None:
                     self.do_done(who, what, multiple=multiple)
@@ -1139,7 +1213,7 @@ class ADVCharacter(object):
                     self.do_done(who, what)
 
                 # Finally, log this line of dialogue.
-                if who and isinstance(who, (str, unicode)):
+                if who and isinstance(who, basestring):
                     renpy.exports.log(who)
 
                 renpy.exports.log(what)
@@ -1149,19 +1223,30 @@ class ADVCharacter(object):
 
             if (multiple is None) and interact:
                 renpy.store._side_image_attributes = old_side_image_attributes
-                self.restore_say_attributes(False, old_attr_state, interact)
+
+                if old_attr_state is not None:
+                    _, images = old_attr_state
+                    before = images.get_attributes(None, self.image_tag)
+
+                if self.restore_say_attributes(False, old_attr_state, interact):
+                    after = images.get_attributes(None, self.image_tag)
+                    self.handle_say_transition('restore', before, after)
+
+    @property
+    def statement_name(self):
+        if not (self.condition is None or renpy.python.py_eval(self.condition)):
+            return "say-condition-false"
+        else:
+            return "say"
 
     def predict(self, what):
 
         old_attr_state = self.handle_say_attributes(True, True)
 
-        if renpy.config.speaking_attribute is not None:
-            self.resolve_say_attributes(True, wanted=[ renpy.config.speaking_attribute ])
-
         old_side_image_attributes = renpy.store._side_image_attributes
 
         if self.image_tag:
-            attrs = ( self.image_tag, ) + renpy.game.context().images.get_attributes("master", self.image_tag)
+            attrs = (self.image_tag,) + renpy.game.context().images.get_attributes("master", self.image_tag)
         else:
             attrs = None
 
@@ -1198,10 +1283,10 @@ class ADVCharacter(object):
         if history_length is None:
             return
 
-        if not renpy.store._history:  # @UndefinedVariable
+        if not renpy.store._history: # @UndefinedVariable
             return
 
-        history = renpy.store._history_list  # @UndefinedVariable
+        history = renpy.store._history_list # @UndefinedVariable
 
         h = HistoryEntry()
 
@@ -1246,10 +1331,12 @@ class ADVCharacter(object):
         if history_length is None:
             return
 
-        if not renpy.store._history:  # @UndefinedVariable
+        if not renpy.store._history: # @UndefinedVariable
             return
 
-        renpy.store._history_list.pop()  # @UndefinedVariable
+        # The history can be reset at any time, so check that we have some.
+        if renpy.store._history_list:
+            renpy.store._history_list.pop() # @UndefinedVariable
 
 
 def Character(name=NotSet, kind=None, **properties):
@@ -1317,9 +1404,13 @@ def Character(name=NotSet, kind=None, **properties):
     These options help to control the display of the name.
 
     `dynamic`
-        If true, then `name` should be a string containing a Python
-        expression. That string will be evaluated before each line
-        of dialogue, and the result used as the name of the character.
+        If true, then `name` should either be a string containing a Python
+        expression, a function, or a callable object. If it's a string,
+        That string will be evaluated before each line of dialogue, and
+        the result used as the name of the character. Otherwise, the
+        function or callable object will be called with no arguments
+        before each line of dialogue, and the return value of the call will
+        be used as the name of the character.
 
     **Controlling Interactions.**
     These options control if the dialogue is displayed, if an
