@@ -1,4 +1,4 @@
-# Copyright 2004-2015 Tom Rothamel <pytom@bishoujo.us>
+# Copyright 2004-2021 Tom Rothamel <pytom@bishoujo.us>
 #
 # Permission is hereby granted, free of charge, to any person
 # obtaining a copy of this software and associated documentation files
@@ -24,6 +24,9 @@
 # are None, at least to the point of making it through __init__. This is
 # so that prediction of images works.
 
+from __future__ import division, absolute_import, with_statement, print_function, unicode_literals
+from renpy.compat import *
+
 import renpy.display
 from renpy.display.render import render
 
@@ -63,6 +66,7 @@ def null_render(d, width, height, st, at):
 
     return rv
 
+
 class NoTransition(Transition):
     """
     :doc: transition function
@@ -108,7 +112,7 @@ class MultipleTransition(Transition):
     scene following it. For example::
 
         define logodissolve = MultipleTransition([
-            False, Dissolve(0.5)
+            False, Dissolve(0.5),
             "logo.jpg", Pause(1.0),
             "logo.jpg", dissolve,
             True])
@@ -196,7 +200,6 @@ def Fade(out_time,
          widget=None,
          alpha=False,
          ):
-
     """
     :doc: transition function
     :args: (out_time, hold_time, in_time, color="#000")
@@ -290,7 +293,6 @@ class Pixellate(Transition):
             visible = self.new_widget
             self.events = True
 
-
         rdr = render(visible, width, height, st, at)
         rv = renpy.display.render.Render(rdr.width, rdr.height)
 
@@ -299,6 +301,12 @@ class Pixellate(Transition):
         rv.operation = renpy.display.render.PIXELLATE
         rv.operation_parameter = 2 ** step
 
+        rv.mesh = True
+        rv.add_shader("renpy.texture")
+        rv.add_property("texture_scaling", "nearest_mipmap_nearest")
+        rv.add_property("anisotropic", False)
+        rv.add_uniform("u_lod_bias", step + 1)
+
         renpy.display.render.redraw(self, 0)
 
         return rv
@@ -316,9 +324,7 @@ class Dissolve(Transition):
         The time the dissolve will take.
 
     `alpha`
-        If true, the dissolve will alpha-composite the the result of the
-        transition with the screen. If false, the result of the transition
-        will replace the screen, which is more efficient.
+        Ignored.
 
     `time_warp`
         A function that adjusts the timeline. If not None, this should be a
@@ -344,7 +350,6 @@ class Dissolve(Transition):
         self.alpha = alpha
         self.time_warp = time_warp
 
-
     def render(self, width, height, st, at):
 
         if renpy.game.less_updates:
@@ -365,12 +370,26 @@ class Dissolve(Transition):
         width = min(top.width, bottom.width)
         height = min(top.height, bottom.height)
 
-        rv = renpy.display.render.Render(width, height, opaque=not self.alpha)
+        rv = renpy.display.render.Render(width, height, opaque=not (self.alpha or renpy.config.dissolve_force_alpha))
 
         rv.operation = renpy.display.render.DISSOLVE
-        rv.operation_alpha = self.alpha
+        rv.operation_alpha = self.alpha or renpy.config.dissolve_force_alpha
         rv.operation_complete = complete
 
+        if renpy.display.render.models:
+
+            target = rv.get_size()
+
+            if top.get_size() != target:
+                top = top.subsurface((0, 0, width, height))
+            if bottom.get_size() != target:
+                bottom = bottom.subsurface((0, 0, width, height))
+
+            rv.mesh = True
+            rv.add_shader("renpy.dissolve")
+            rv.add_uniform("u_renpy_dissolve", complete)
+            rv.add_property("mipmap", renpy.config.mipmap_dissolves if (self.style.mipmap is None) else self.style.mipmap)
+
         rv.blit(bottom, (0, 0), focus=False, main=False)
         rv.blit(top, (0, 0), focus=True, main=True)
 
@@ -404,12 +423,10 @@ class ImageDissolve(Transition):
         completed one step of dissolving in.
 
     `reverse`
-        If true, black pixels will dissolve in before white pixels.
+        If True, black pixels will dissolve in before white pixels.
 
     `alpha`
-        If true, the dissolve will alpha-composite the the result of the
-        transition with the screen. If false, the result of the transition
-        will replace the screen, which is more efficient.
+        Ignored.
 
     `time_warp`
         A function that adjusts the timeline. If not None, this should be a
@@ -432,18 +449,18 @@ class ImageDissolve(Transition):
     time_warp = None
 
     def __init__(
-        self,
-        image,
-        time,
-        ramplen=8,
-        ramptype='linear',
-        ramp=None,
-        reverse=False,
-        alpha=False,
-        old_widget=None,
-        new_widget=None,
-        time_warp=None,
-        **properties):
+            self,
+            image,
+            time,
+            ramplen=8,
+            ramptype='linear',
+            ramp=None,
+            reverse=False,
+            alpha=False,
+            old_widget=None,
+            new_widget=None,
+            time_warp=None,
+            **properties):
 
         # ramptype and ramp are now unused, but are kept for compatbility with
         # older code.
@@ -472,7 +489,7 @@ class ImageDissolve(Transition):
                 0, 0, 0, 0, 1,
                 0, 0, 0, 0, 1,
                 0, 0, 0, 0, 1,
-                - 1, 0, 0, 0, 1)
+                -1, 0, 0, 0, 1)
 
         self.image = renpy.display.im.MatrixColor(image, matrix)
 
@@ -482,11 +499,9 @@ class ImageDissolve(Transition):
         # The length of the ramp.
         self.ramplen = max(ramplen, 1)
 
-
     def visit(self):
         return super(ImageDissolve, self).visit() + [ self.image ]
 
-
     def render(self, width, height, st, at):
 
         if renpy.game.less_updates or renpy.display.less_imagedissolve:
@@ -503,7 +518,7 @@ class ImageDissolve(Transition):
         width = min(bottom.width, top.width, image.width)
         height = min(bottom.height, top.height, image.height)
 
-        rv = renpy.display.render.Render(width, height, opaque=not self.alpha)
+        rv = renpy.display.render.Render(width, height, opaque=not (self.alpha or renpy.config.dissolve_force_alpha))
 
         complete = st / self.delay
 
@@ -511,10 +526,39 @@ class ImageDissolve(Transition):
             complete = self.time_warp(complete)
 
         rv.operation = renpy.display.render.IMAGEDISSOLVE
-        rv.operation_alpha = self.alpha
+        rv.operation_alpha = self.alpha or renpy.config.dissolve_force_alpha
         rv.operation_complete = complete
         rv.operation_parameter = self.ramplen
 
+        if renpy.display.render.models:
+
+            target = rv.get_size()
+
+            if image.get_size() != target:
+                image = image.subsurface((0, 0, width, height))
+            if top.get_size() != target:
+                top = top.subsurface((0, 0, width, height))
+            if bottom.get_size() != target:
+                bottom = bottom.subsurface((0, 0, width, height))
+
+            ramp = self.ramplen
+
+            # Prevent a DBZ if the user gives us a 0 ramp.
+            if ramp < 1:
+                ramp = 1
+
+            # Compute the offset to apply to the alpha.
+            start = -1.0
+            end = ramp / 256.0
+            offset = start + (end - start) * complete
+
+            rv.mesh = True
+
+            rv.add_shader("renpy.imagedissolve",)
+            rv.add_uniform("u_renpy_dissolve_offset", offset)
+            rv.add_uniform("u_renpy_dissolve_multiplier", 256.0 / ramp)
+            rv.add_property("mipmap", renpy.config.mipmap_dissolves if (self.style.mipmap is None) else self.style.mipmap)
+
         rv.blit(image, (0, 0), focus=False, main=False)
         rv.blit(bottom, (0, 0), focus=False, main=False)
         rv.blit(top, (0, 0), focus=True, main=True)
@@ -541,8 +585,7 @@ class AlphaDissolve(Transition):
         The time the transition takes, before ending.
 
     `alpha`
-        If true, the image is composited with what's behind it. If false,
-        the default, the image is opaque and overwrites what's behind it.
+        Ignored.
 
     `reverse`
         If true, the alpha channel is reversed. Opaque areas are taken
@@ -550,15 +593,17 @@ class AlphaDissolve(Transition):
         new image.
      """
 
+    mipmap = None
+
     def __init__(
-        self,
-        control,
-        delay=0.0,
-        old_widget=None,
-        new_widget=None,
-        alpha=False,
-        reverse=False,
-        **properties):
+            self,
+            control,
+            delay=0.0,
+            old_widget=None,
+            new_widget=None,
+            alpha=False,
+            reverse=False,
+            **properties):
 
         super(AlphaDissolve, self).__init__(delay, **properties)
 
@@ -594,10 +639,17 @@ class AlphaDissolve(Transition):
         rv = renpy.display.render.Render(width, height, opaque=not self.alpha)
 
         rv.operation = renpy.display.render.IMAGEDISSOLVE
-        rv.operation_alpha = self.alpha
+        rv.operation_alpha = self.alpha or renpy.config.dissolve_force_alpha
         rv.operation_complete = 256.0 / (256.0 + 256.0)
         rv.operation_parameter = 256
 
+        if renpy.display.render.models:
+            rv.mesh = True
+            rv.add_shader("renpy.imagedissolve",)
+            rv.add_uniform("u_renpy_dissolve_offset", 0)
+            rv.add_uniform("u_renpy_dissolve_multiplier", 1.0)
+            rv.add_property("mipmap", renpy.config.mipmap_dissolves if (self.style.mipmap is None) else self.style.mipmap)
+
         rv.blit(control, (0, 0), focus=False, main=False)
 
         if not self.reverse:
@@ -618,7 +670,7 @@ class CropMove(Transition):
 
     Returns a transition that works by cropping a scene and positioning it on the
     screen. This can be used to implement a variety of effects, all of which
-    involved changing rectangular slices of scenes.
+    involve changing rectangular slices of scenes.
 
     `time`
         The time the transition takes.
@@ -804,7 +856,6 @@ class CropMove(Transition):
             endcrop = (0.5, 0.5, 0.0, 0.0)
             topnew = False
 
-
         elif mode == "custom":
             pass
         else:
@@ -855,7 +906,6 @@ class CropMove(Transition):
         crop = interpolate_tuple(self.startcrop, self.endcrop)
         pos = interpolate_tuple(self.startpos, self.endpos)
 
-
         top = render(self.top, width, height, st, at)
         bottom = render(self.bottom, width, height, st, at)
 
@@ -872,6 +922,131 @@ class CropMove(Transition):
         return rv
 
 
+class PushMove(Transition):
+    """
+    :doc: transition function
+    :args: (time, mode="pushright")
+    :name: PushMove
+
+    Returns a transition that works by taking the new scene and using it to
+    "push" the old scene off the screen.
+
+    `time`
+        The time the transition takes.
+
+    `mode`
+        There are four possible modes: "pushright", "pushleft", "pushup",
+        and "pushdown", which push the old scene off the screen in the
+        direction indicated.
+
+    ::
+
+        define pushright = PushMove(1.0, "pushright")
+        define pushleft = PushMove(1.0, "pushleft")
+        define pushup = PushMove(1.0, "pushup")
+        define pushdown = PushMove(1.0, "pushdown")
+    """
+
+    def __init__(self, time,
+                 mode="pushright",
+                 old_widget=None,
+                 new_widget=None,
+                 **properties):
+
+        super(PushMove, self).__init__(time, **properties)
+        self.time = time
+
+        if mode == "pushright":
+            self.new_startpos = (0.0, 0.0)
+            self.new_startcrop = (1.0, 0.0, 0.0, 1.0)
+            self.new_endpos = (0.0, 0.0)
+            self.new_endcrop = (0.0, 0.0, 1.0, 1.0)
+            self.old_endpos = (1.0, 0.0)
+            self.old_endcrop = (0.0, 0.0, 0.0, 1.0)
+            self.old_startpos = (0.0, 0.0)
+            self.old_startcrop = (0.0, 0.0, 1.0, 1.0)
+
+        elif mode == "pushleft":
+            self.new_startpos = (1.0, 0.0)
+            self.new_startcrop = (0.0, 0.0, 0.0, 1.0)
+            self.new_endpos = (0.0, 0.0)
+            self.new_endcrop = (0.0, 0.0, 1.0, 1.0)
+            self.old_endpos = (0.0, 0.0)
+            self.old_endcrop = (1.0, 0.0, 0.0, 1.0)
+            self.old_startpos = (0.0, 0.0)
+            self.old_startcrop = (0.0, 0.0, 1.0, 1.0)
+
+        elif mode == "pushup":
+            self.new_startpos = (0.0, 1.0)
+            self.new_startcrop = (0.0, 0.0, 1.0, 0.0)
+            self.new_endpos = (0.0, 0.0)
+            self.new_endcrop = (0.0, 0.0, 1.0, 1.0)
+            self.old_endpos = (0.0, 0.0)
+            self.old_endcrop = (0.0, 1.0, 1.0, 0.0)
+            self.old_startpos = (0.0, 0.0)
+            self.old_startcrop = (0.0, 0.0, 1.0, 1.0)
+
+        elif mode == "pushdown":
+            self.new_startpos = (0.0, 0.0)
+            self.new_startcrop = (0.0, 1.0, 1.0, 0.0)
+            self.new_endpos = (0.0, 0.0)
+            self.new_endcrop = (0.0, 0.0, 1.0, 1.0)
+            self.old_endpos = (0.0, 1.0)
+            self.old_endcrop = (0.0, 0.0, 1.0, 0.0)
+            self.old_startpos = (0.0, 0.0)
+            self.old_startcrop = (0.0, 0.0, 1.0, 1.0)
+
+        else:
+            raise Exception("Invalid mode %s passed into PushMove." % mode)
+
+        self.delay = time
+        self.time = time
+
+        self.old_widget = old_widget
+        self.new_widget = new_widget
+
+        self.events = False
+
+    def render(self, width, height, st, at):
+
+        if renpy.game.less_updates:
+            return null_render(self, width, height, st, at)
+
+        time = 1.0 * st / self.time
+
+        # Done rendering.
+        if time >= 1.0:
+            self.events = True
+            return render(self.new_widget, width, height, st, at)
+
+        # How we scale each element of a tuple.
+        scales = (width, height, width, height)
+
+        def interpolate_tuple(t0, t1):
+            return tuple([ int(s * (a * (1.0 - time) + b * time))
+                           for a, b, s in zip(t0, t1, scales) ])
+
+        new_crop = interpolate_tuple(self.new_startcrop, self.new_endcrop)
+        new_pos = interpolate_tuple(self.new_startpos, self.new_endpos)
+
+        old_crop = interpolate_tuple(self.old_startcrop, self.old_endcrop)
+        old_pos = interpolate_tuple(self.old_startpos, self.old_endpos)
+
+        new = render(self.new_widget, width, height, st, at)
+        old = render(self.old_widget, width, height, st, at)
+
+        rv = renpy.display.render.Render(width, height)
+
+        old_ss = old.subsurface(old_crop, focus=True)
+        rv.blit(old_ss, old_pos, focus=True)
+
+        new_ss = new.subsurface(new_crop, focus=True)
+        rv.blit(new_ss, new_pos, focus=True)
+
+        renpy.display.render.redraw(self, 0)
+        return rv
+
+
 def ComposeTransition(trans, before=None, after=None, new_widget=None, old_widget=None):
     """
     :doc: transition function
@@ -920,4 +1095,3 @@ def SubTransition(rect, trans, old_widget=None, new_widget=None, **properties):
     f.add(inner)
 
     return NoTransition(delay, old_widget=f, new_widget=f)
-
